from __future__ import absolute_import

import collections
import urllib

import tornado.web
import tornado.httpclient

try:
    from tornado.curl_httpclient import CurlAsyncHTTPClient as AsyncHTTPClient
except ImportError:
    from tornado.httpclient import AsyncHTTPClient

try:
    import ujson as json
except ImportError:
    import json


#region Internal


http_client = AsyncHTTPClient(max_clients=1024)


def json2namedtuple(obj):
    if isinstance(obj, dict):
        keys = []
        values = []
        for key, value in obj.iteritems():
            keys.append(key)
            values.append(json2namedtuple(value))
        return collections.namedtuple("object", keys)(*values)
    elif isinstance(obj, (set, list, tuple)):
        array = []
        for item in obj:
            array.append(json2namedtuple(item))
        return array
    else:
        return obj

#endregion


#region Public

class Client(object):
    def __init__(self, url):
        self.url = url.strip("/")
        self.item = ItemRepository(self)
        self.news = NewsRepository(self)

    def request(self, verb, path, data=None, callback=None):
        def _callback(response):
            if response.code == 200:
                callback(json2namedtuple(json.loads(response.body)) if response.body else None)
            elif response.code == 404:
                callback(None)
            else:
                raise tornado.httpclient.HTTPError(response.code, response.effective_url)

        url = "%s/%s" % (self.url, path)
        if verb == "GET":
            url += "?%s" % urllib.urlencode(dict((k, v) for k, v in data.iteritems() if v is not None), True)
            data = None
        else:
            if data:
                data = json.dumps(data)

        http_client.fetch(tornado.httpclient.HTTPRequest(url, method=verb, body=data), _callback)

    def get(self, path, callback, **query_args):
        self.request("GET", path, query_args, callback)

    def pipeline(self):
        return ClientPipeline(self)


class ClientPipeline(Client):
    def __init__(self, client):
        super(ClientPipeline, self).__init__(client.url)
        self._pipe = []

    def request(self, verb, path, data=None, callback=None):
        self._pipe.append({"path": path, "args": data})

    def execute(self, callback):
        super(ClientPipeline, self).request("POST", "", self._pipe, callback)


class Repository(object):
    def __init__(self, client):
        self.client = client


class ItemRepository(Repository):
    def get(self, item_id, language=None, callback=None):
        self.client.get("item/get", item_id=item_id, language=language, callback=callback)

    def search(self, query="", skip=0, take=10, language=None, callback=None):
        self.client.get("item/search", query=query, skip=skip, take=take, language=language, callback=callback)

    def list(self, item_ids, language=None, callback=None):
        self.client.get("item/list", _item_ids=item_ids, language=language, callback=callback)


class NewsRepository(Repository):
    def get(self, article_id):
        pass

    def search(self):
        pass


#endregion